package rabbit.client;

import java.io.Closeable;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URL;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.concurrent.atomic.AtomicInteger;
import rabbit.http.HttpHeader;
import rabbit.httpio.BlockListener;
import rabbit.httpio.ResourceSource;
import rabbit.io.BufferHandle;
import rabbit.io.WebConnection;
import rabbit.util.Logger;

/** A class to download a resource. 
 *  This is mostly an example of how to use the rabbit client classes.
 *
 * @author <a href="mailto:robo@khelekore.org">Robert Olofsson</a>
 */
public class WGet {
    private AtomicInteger outstandingRequests = new AtomicInteger ();
    private ClientBase clientBase;
    private BlockHandler blockHandler;

    public static void main (String[] args) {
	try {
	    WGet wget = new WGet ();
	    wget.get (args);
	} catch (IOException e) {
	    e.printStackTrace ();
	}
    }

    public WGet () throws IOException {
	clientBase = new ClientBase ();
    }

    public void get (String[] urls) throws IOException {
	for (String url : urls)
	    get (url);
    }

    public void get (String url) throws IOException {
	outstandingRequests.incrementAndGet ();
	clientBase.sendRequest (clientBase.getRequest ("GET", url),
				new WGetListener (clientBase.getLogger ()));
    }

    private void decrementOutstanding () {
	int outstanding = outstandingRequests.decrementAndGet ();
	if (outstanding == 0)
	    clientBase.shutdown ();
    }

    private void close (Closeable c) {
	if (c == null)
	    return;
	try {
	    c.close ();
	} catch (IOException e) {
	    Logger l = clientBase.getLogger ();
	    l.logError ("Failed to close connection: " + c);
	}
    }

    private class WGetListener extends ClientListenerAdapter {   
	public WGetListener (Logger logger) {
	    super (logger);
	}
	
	@Override public void redirectedTo (String url) throws IOException {
	    get (url);
	}

	@Override public void handleResponse (HttpHeader request, 
					      HttpHeader response, 
					      WebConnection wc,
					      BufferHandle bufferHandle, 
					      boolean keepalive, 
					      boolean isChunked,
					      long dataSize) {
	    ResourceSource wrs = 
		clientBase.getWebConnectionResouceSource (wc, bufferHandle, 
							  isChunked, dataSize);
	    try {
		blockHandler = new BlockHandler (request, wc, wrs);
		wrs.addBlockListener (blockHandler);
	    } catch (IOException e) {
		close (wc);
		handleFailure (request, e);
	    }
	}

	@Override public void requestDone () {
	    decrementOutstanding ();
	}
    }

    private class BlockHandler implements BlockListener {
	private WebConnection wc;
	private ResourceSource rs;
	private FileChannel fc;

	public BlockHandler (HttpHeader request, WebConnection wc, 
			     ResourceSource rs) 
	    throws IOException {
	    URL u = new URL (request.getRequestURI ());
	    File f = new File (getFileName (u));
	    if (f.exists ())
		throw new IOException ("File already exists: " + f.getName ());
	    FileOutputStream fos = new FileOutputStream (f);
	    fc = fos.getChannel ();
	    this.wc = wc;
	    this.rs = rs;
	}

	private String getFileName (URL u) {
	    String s = u.getFile ();
	    int i = s.lastIndexOf ('/');
	    if (i > -1) 
		s = s.substring (i + 1);
	    if (s.equals (""))
		return "index.html";
	    return s;
	}

	public void bufferRead (final BufferHandle bufHandle) {
	    clientBase.getTaskRunner ().runThreadTask (new Runnable () {
		    public void run () {
			try {
			    ByteBuffer buf = bufHandle.getBuffer ();
			    fc.write (buf);
			    readMore ();
			} catch (IOException e) {
			    failed (e);
			}			
		    }
		});
	}

	private void readMore () {
	    clientBase.getTaskRunner ().runMainTask (new Runnable () {
		    public void run () {
			rs.addBlockListener (BlockHandler.this);
		    }
		});
	}

	public void finishedRead () {
	    clientBase.getConnectionHandler ().releaseConnection (wc);
	    close (fc);
	    decrementOutstanding ();
	}
	
	public void failed (Exception cause) {
	    Logger l = clientBase.getLogger ();
	    l.logError ("Failed to download resource");
	    downloadFailed ();
	}
	
	public void timeout () {
	    Logger l = clientBase.getLogger ();
	    l.logError ("Timed out trying to download resource");
	}

	private void downloadFailed () {
	    close (wc);
	    close (fc);
	    decrementOutstanding ();
	}
    }
}
